/* #includes */ /*{{{C}}}*//*{{{*/
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "hgc.h"
/*}}}*/

/* variables */ /*{{{*/
static struct termios tty;
static int ttyfd;
static int reverse=0;
/*}}}*/

/* raw */ /*{{{*/
static void raw(void)
{
  struct termios rtty;

  tcgetattr(ttyfd,&tty);
  rtty=tty;
  memset(&(rtty.c_cc[0]),255,NCCS);
  rtty.c_cc[VMIN]=1;
  rtty.c_cc[VTIME]=0;
  rtty.c_iflag=0;
  rtty.c_oflag=0;
  rtty.c_lflag=0;
  tcsetattr(ttyfd,TCSANOW,&rtty);
}
/*}}}*/
/* cooked */ /*{{{*/
static void cooked(void)
{
  tcsetattr(ttyfd,TCSANOW,&tty);
}
/*}}}*/
/* die */ /*{{{*/
static void die(int n)
{
  textmode();
  cooked();
  fprintf(stderr,"hgcpbm: terminating due to signal %d.\n",n);
  exit(1);
}
/*}}}*/
/* readpbm */ /*{{{*/
static unsigned char *readpbm(char *name,int *width,int *height)
{
  /* variables */ /*{{{*/
  unsigned char *pbm,*pbmrun;
  FILE *fp;
  char magic[10];
  char size[256];
  int i;
  /*}}}*/

  if (!strcmp(name,"-")) fp=stdin;
  else if ((fp=fopen(name,"rb"))==(FILE*)0)
  {
    pout="can't open file";
    return (unsigned char*)0;
  }
  if (fgets(magic,sizeof(magic),fp)==(char*)0 || magic[0]!='P' || magic[1]!='4')
  {
    pout="bad magic number";
    fclose(fp);
    return (unsigned char *)0;
  }
  while (fgets(size,sizeof(size),fp))
  {
    if (sscanf(size,"%d %d",width,height)!=2 && *size!='#')
    {
      fprintf(stderr,"<%s>\n",size);
      pout="no width and height";
      fclose(fp);
      return (unsigned char *)0;
    }
    if (*size!='#') break;
  }
  pbm=malloc(((*width+7)>>3)**height);
  if (fread(pbm,((*width+7)>>3)**height,1,fp)!=1)
  {
    free(pbm);
    pout="unexpected eof";
    return (unsigned char*)0;
  }
  pbmrun=pbm; if (reverse) for (i=0; i<(*height*((*width+7)>>3)); pbmrun++,i++) *pbmrun=~*pbmrun;
  return pbm;
}
/*}}}*/

/* main */ /*{{{*/
int main(int argc, char *argv[])
{
  /* variables */ /*{{{*/
  unsigned char *vmem,*pbm;
  int width,height,y,bytes_wide,x_off,y_off,max_x_off,max_y_off;
  char c;
  /*}}}*/

  /* parse arguments */ /*{{{*/
  {
    int c;
    int usage=0;

    while ((c=getopt(argc,argv,"r"))!=EOF)
    {
      switch (c)
      {
        case 'r': reverse=1; break;
        case '?':
        default: usage=1; break;
      }
    }

    if (usage || optind>=argc)
    {
      fprintf(stderr,"Usage: hgcpbm [-r] file\n");
      exit(1);
    }
  }
  /*}}}*/
  /* map screen to memory */ /*{{{*/
  if ((vmem=mapscreen())==(unsigned char*)0)
  {
    fprintf(stderr,"hgcpbm: %s: %s\n",pout,strerror(errno));
    exit(1);
  }
  if ((ttyfd=open("/dev/tty",O_RDONLY))<0)
  {
    fprintf(stderr,"hgcpbm: can't open tty: %s\n",strerror(errno));
    exit(1);
  }
  /*}}}*/
  /* set up signals */ /*{{{*/
  signal(SIGINT,die);
  signal(SIGTERM,die);
  signal(SIGSEGV,die);
  /*}}}*/
  /* read pbm file */ /*{{{*/
  if ((pbm=readpbm(argv[optind],&width,&height))==(unsigned char*)0)
  {
    fprintf(stderr,"hgcpbm: can't read pbm file %s: %s\n",argv[optind],pout);
    exit(1);
  }
  /*}}}*/
  raw();
  grafmode();
  bytes_wide=((width+7)>>3);
  x_off=0; max_x_off=(width>720 ? width-720 : 0);
  y_off=0; max_y_off=(height>348 ? height-348 : 0);
  do
  {
    /* display window of picture */ /*{{{*/
    for (y=0; y<348; y++)
    {
      unsigned char *start=vmem+((y&3)<<13)+(y>>2)*90;
    
      if (bytes_wide<90) memset(start+bytes_wide,0,90-bytes_wide);
      if (y<height) memcpy(start,pbm+((x_off+7)>>3)+(y+y_off)*bytes_wide,bytes_wide>90 ? 90 : bytes_wide);
      else memset(start,0,bytes_wide);
    }
    /*}}}*/
    /* read and process key */ /*{{{*/
    if (read(ttyfd,&c,1)!=1) c='q';
    switch (c)
    {
      case 'l': x_off+=8; if (x_off>max_x_off) x_off=max_x_off; break;
      case 'h': x_off-=8; if (x_off<0) x_off=0; break;
      case 'j': y_off+=8; if (y_off>max_y_off) y_off=max_y_off; break;
      case 'k': y_off-=8; if (y_off<0) y_off=0; break;
      case '\033': if (read(ttyfd,&c,1)!=1) c='q'; else if (c=='[')
      {
        if (read(ttyfd,&c,1)!=1) c='q';
        switch (c)
        {
          case 'C': x_off+=8; if (x_off>max_x_off) x_off=max_x_off; break;
          case 'D': x_off-=8; if (x_off<0) x_off=0; break;
          case 'B': y_off+=8; if (y_off>max_y_off) y_off=max_y_off; break;
          case 'A': y_off-=8; if (y_off<0) y_off=0; break;
        }
      }
    }
    /*}}}*/
  } while (c!='q');
  textmode();
  cooked();
  return 0;
}
/*}}}*/
